@@ -577,6 +577,10 @@ ftp_parse_vms_ls (const char *file)
     }
   dir = l = NULL;
 
+  /* #### The next three lines are a memory leak because they don't
+     bother to free the pointer that read_whole_line() returns!
+     FIXME! */
+
   /* Empty line */
   read_whole_line (fp);
   /* "Directory PUB$DEVICE[PUB]" */
@@ -796,7 +800,7 @@ Unsupported listing type, trying Unix listing parser.\n"));
    directories and files on the appropriate host.  The references are
    FTP.  */
 uerr_t
-ftp_index (const char *file, struct urlinfo *u, struct fileinfo *f)
+ftp_index (const char *file, struct url *u, struct fileinfo *f)
 {
   FILE *fp;
   char *upwd;
