@@ -1,5 +1,5 @@
 /* Declarations for host.c
-   Copyright (C) 1995, 1996, 1997 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1997, 2001 Free Software Foundation, Inc.
 
 This file is part of GNU Wget.
 
@@ -20,22 +20,28 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #ifndef HOST_H
 #define HOST_H
 
-struct urlinfo;
+struct url;
+struct address_list;
 
 /* Function declarations */
-
-struct hostent *ngethostbyname PARAMS ((const char *));
-int store_hostaddress PARAMS ((unsigned char *, const char *));
-
-void clean_hosts PARAMS ((void));
-
-char *realhost PARAMS ((const char *));
-int same_host PARAMS ((const char *, const char *));
-int accept_domain PARAMS ((struct urlinfo *));
-int sufmatch PARAMS ((const char **, const char *));
-
-char *ftp_getaddress PARAMS ((void));
-
+struct address_list *lookup_host PARAMS ((const char *, int));
 char *herrmsg PARAMS ((int));
 
+void address_list_get_bounds PARAMS ((struct address_list *, int *, int *));
+void address_list_copy_one PARAMS ((struct address_list *, int,
+				    unsigned char *));
+int address_list_match_all PARAMS ((struct address_list *,
+				    struct address_list *));
+void address_list_set_faulty PARAMS ((struct address_list *, int));
+void address_list_release PARAMS ((struct address_list *));
+
+/* This was originally going to be a macro, but then every caller
+   would have to #include the netinet stuff.  */
+char *pretty_print_address PARAMS ((const unsigned char *));
+
+int accept_domain PARAMS ((struct url *));
+int sufmatch PARAMS ((const char **, const char *));
+
+void host_cleanup PARAMS ((void));
+
 #endif /* HOST_H */
