@@ -432,7 +432,9 @@ advance_declaration (const char *beg, const char *end)
 	    state = AC_S_DEFAULT;
 	  break;
 	case AC_S_QUOTE1:
-	  assert (ch == '\'' || ch == '"');
+	  /* We must use 0x22 because broken assert macros choke on
+	     '"' and '\"'.  */
+	  assert (ch == '\'' || ch == 0x22);
 	  quote_char = ch;	/* cheating -- I really don't feel like
 				   introducing more different states for
 				   different quote characters. */
@@ -638,6 +640,19 @@ map_html_tags (const char *text, int size,
 
 	SKIP_WS (p);
 
+	if (*p == '/')
+	  {
+	    /* A slash at this point means the tag is about to be
+	       closed.  This is legal in XML and has been popularized
+	       in HTML via XHTML.  */
+	    /* <foo a=b c=d /> */
+	    /*              ^  */
+	    ADVANCE (p);
+	    SKIP_WS (p);
+	    if (*p != '>')
+	      goto backout_tag;
+	  }
+
 	/* Check for end of tag definition. */
 	if (*p == '>')
 	  break;
@@ -654,7 +669,7 @@ map_html_tags (const char *text, int size,
 
 	/* Establish bounds of attribute value. */
 	SKIP_WS (p);
-	if (NAME_CHAR_P (*p) || *p == '>')
+	if (NAME_CHAR_P (*p) || *p == '/' || *p == '>')
 	  {
 	    /* Minimized attribute syntax allows `=' to be omitted.
                For example, <UL COMPACT> is a valid shorthand for <UL
@@ -735,7 +750,7 @@ map_html_tags (const char *text, int size,
 	    /* We skipped the whitespace and found something that is
 	       neither `=' nor the beginning of the next attribute's
 	       name.  Back out.  */
-	    goto backout_tag;	/* <foo bar /... */
+	    goto backout_tag;	/* <foo bar [... */
 				/*          ^    */
 	  }
 
