@@ -45,12 +45,20 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
 #include "wget.h"
 #include "utils.h"
-#include "getopt.h"
 #include "init.h"
 #include "retr.h"
 #include "recur.h"
 #include "host.h"
 #include "cookies.h"
+#include "url.h"
+#include "progress.h"		/* for progress_handle_sigwinch */
+
+#ifdef HAVE_SSL
+# include "gen_sslfunc.h"
+#endif
+
+/* On GNU system this will include system-wide getopt.h. */
+#include "getopt.h"
 
 #ifndef PATH_SEPARATOR
 # define PATH_SEPARATOR '/'
@@ -100,18 +108,6 @@ i18n_initialize (void)
   textdomain ("wget");
 #endif /* HAVE_NLS */
 }
-
-/* It's kosher to declare these here because their interface _has_ to
-   be void foo(void).  */
-void host_init PARAMS ((void));
-
-/* This just calls the various initialization functions from the
-   modules that need one-time initialization. */
-static void
-private_initialize (void)
-{
-  host_init ();
-}
 
 /* Print the usage message.  */
 static void
@@ -154,6 +150,7 @@ Logging and input file:\n\
   -B,  --base=URL             prepends URL to relative links in -F -i file.\n\
        --sslcertfile=FILE     optional client certificate.\n\
        --sslcertkey=KEYFILE   optional keyfile for this certificate.\n\
+       --egd-file=FILE        file name of the EGD socket.\n\
 \n"), stdout);
   fputs (_("\
 Download:\n\
@@ -162,15 +159,17 @@ Download:\n\
   -O   --output-document=FILE   write documents to FILE.\n\
   -nc, --no-clobber             don\'t clobber existing files or use .# suffixes.\n\
   -c,  --continue               resume getting a partially-downloaded file.\n\
-       --dot-style=STYLE        set retrieval display style.\n\
+       --progress=TYPE          select progress gauge type.\n\
   -N,  --timestamping           don\'t re-retrieve files unless newer than local.\n\
   -S,  --server-response        print server response.\n\
        --spider                 don\'t download anything.\n\
   -T,  --timeout=SECONDS        set the read timeout to SECONDS.\n\
   -w,  --wait=SECONDS           wait SECONDS between retrievals.\n\
        --waitretry=SECONDS      wait 1...SECONDS between retries of a retrieval.\n\
+       --random-wait            wait from 0...2*WAIT secs between retrievals.\n\
   -Y,  --proxy=on/off           turn proxy on or off.\n\
   -Q,  --quota=NUMBER           set retrieval quota to NUMBER.\n\
+       --limit-rate=RATE        limit download rate to RATE.\n\
 \n"), stdout);
   fputs (_("\
 Directories:\n\
@@ -228,7 +227,6 @@ Recursive accept/reject:\n\
   -L,  --relative                   follow relative links only.\n\
   -I,  --include-directories=LIST   list of allowed directories.\n\
   -X,  --exclude-directories=LIST   list of excluded directories.\n\
-  -nh, --no-host-lookup             don\'t DNS-lookup hosts.\n\
   -np, --no-parent                  don\'t ascend to the parent directory.\n\
 \n"), stdout);
   fputs (_("Mail bug reports and suggestions to <bug-wget@gnu.org>.\n"),
@@ -240,7 +238,6 @@ main (int argc, char *const *argv)
 {
   char **url, **t;
   int i, c, nurl, status, append_to_log;
-  int wr = 0;
 
   static struct option long_options[] =
   {
@@ -270,6 +267,7 @@ main (int argc, char *const *argv)
     { "passive-ftp", no_argument, NULL, 139 },
     { "page-requisites", no_argument, NULL, 'p' },
     { "quiet", no_argument, NULL, 'q' },
+    { "random-wait", no_argument, NULL, 165 },
     { "recursive", no_argument, NULL, 'r' },
     { "relative", no_argument, NULL, 'L' },
     { "retr-symlinks", no_argument, NULL, 137 },
@@ -306,10 +304,12 @@ main (int argc, char *const *argv)
     { "include-directories", required_argument, NULL, 'I' },
     { "input-file", required_argument, NULL, 'i' },
     { "level", required_argument, NULL, 'l' },
+    { "limit-rate", required_argument, NULL, 164 },
     { "load-cookies", required_argument, NULL, 161 },
     { "no", required_argument, NULL, 'n' },
     { "output-document", required_argument, NULL, 'O' },
     { "output-file", required_argument, NULL, 'o' },
+    { "progress", required_argument, NULL, 163 },
     { "proxy", required_argument, NULL, 'Y' },
     { "proxy-passwd", required_argument, NULL, 144 },
     { "proxy-user", required_argument, NULL, 143 },
@@ -324,6 +324,7 @@ main (int argc, char *const *argv)
 #ifdef HAVE_SSL
     { "sslcertfile", required_argument, NULL, 158 },
     { "sslcertkey", required_argument, NULL, 159 },
+    { "egd-file", required_argument, NULL, 166 },
 #endif /* HAVE_SSL */
     { "wait", required_argument, NULL, 'w' },
     { "waitretry", required_argument, NULL, 152 },
@@ -331,7 +332,6 @@ main (int argc, char *const *argv)
   };
 
   i18n_initialize ();
-  private_initialize ();
 
   append_to_log = 0;
 
@@ -398,15 +398,15 @@ hpVqvdkKsxmNWrHSLcFbEY:G:g:T:U:O:l:n:i:o:a:t:D:A:R:P:B:e:Q:X:I:w:C:",
 	case 149:
 	  setval ("removelisting", "off");
 	  break;
-	case 150:
-	  setval ("simplehostcheck", "on");
-	  break;
 	case 155:
 	  setval ("bindaddress", optarg);
  	  break;
 	case 156:
 	  setval ("httpkeepalive", "off");
 	  break;
+	case 165:
+	  setval ("randomwait", "on");
+	  break;
 	case 'b':
 	  setval ("background", "on");
 	  break;
@@ -470,7 +470,8 @@ hpVqvdkKsxmNWrHSLcFbEY:G:g:T:U:O:l:n:i:o:a:t:D:A:R:P:B:e:Q:X:I:w:C:",
 	case 'V':
 	  printf ("GNU Wget %s\n\n", version_string);
 	  printf ("%s", _("\
-Copyright (C) 1995, 1996, 1997, 1998, 2000, 2001 Free Software Foundation, Inc.\n\
+Copyright (C) 1995, 1996, 1997, 1998, 2000, 2001 Free Software Foundation, Inc.\n"));
+	  printf ("%s", _("\
 This program is distributed in the hope that it will be useful,\n\
 but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
@@ -515,7 +516,6 @@ GNU General Public License for more details.\n"));
 	  break;
 	case 152:
 	  setval ("waitretry", optarg);
-	  wr = 1;
 	  break;
 	case 153:
 	  setval ("followtags", optarg);
@@ -529,6 +529,12 @@ GNU General Public License for more details.\n"));
 	case 162:
 	  setval ("savecookies", optarg);
 	  break;
+	case 163:
+	  setval ("progress", optarg);
+	  break;
+	case 164:
+	  setval ("limitrate", optarg);
+	  break;
 	case 157:
 	  setval ("referer", optarg);
 	  break;
@@ -539,6 +545,9 @@ GNU General Public License for more details.\n"));
 	case 159:
 	  setval ("sslcertkey", optarg);
 	  break;
+	case 166:
+	  setval ("egdfile", optarg);
+	  break;
 #endif /* HAVE_SSL */
 	case 'A':
 	  setval ("accept", optarg);
@@ -591,7 +600,7 @@ GNU General Public License for more details.\n"));
 	  break;
 	case 'n':
 	  {
-	    /* #### The n? options are utter crock!  */
+	    /* #### What we really want here is --no-foo. */
 	    char *p;
 
 	    for (p = optarg; *p; p++)
@@ -600,9 +609,6 @@ GNU General Public License for more details.\n"));
 		case 'v':
 		  setval ("verbose", "off");
 		  break;
-		case 'h':
-		  setval ("simplehostcheck", "on");
-		  break;
 		case 'H':
 		  setval ("addhostdir", "off");
 		  break;
@@ -673,6 +679,11 @@ GNU General Public License for more details.\n"));
 	}
     }
 
+  /* Initialize progress.  Have to do this after the options are
+     processed so we know where the log file is.  */
+  if (opt.verbose)
+    set_progress_implementation (opt.progress_type);
+
   /* All user options have now been processed, so it's now safe to do
      interoption dependency checks. */
 
@@ -690,18 +701,6 @@ GNU General Public License for more details.\n"));
   if (opt.verbose == -1)
     opt.verbose = !opt.quiet;
 
-  /* Retain compatibility with previous scripts.
-     if wait has been set, but waitretry has not, give it the wait value.
-     A simple check on the values is not enough, I could have set
-     wait to n>0 and waitretry to 0 - HEH */
-  if (opt.wait && !wr)
-    {
-      char  opt_wait_str[256];  /* bigger than needed buf to prevent overflow */
-
-      sprintf(opt_wait_str, "%ld", opt.wait);
-      setval ("waitretry", opt_wait_str);
-    }
-    
   /* Sanity checks.  */
   if (opt.verbose && opt.quiet)
     {
@@ -737,9 +736,14 @@ Can't timestamp and not clobber old files at the same time.\n"));
   /* Fill in the arguments.  */
   for (i = 0; i < nurl; i++, optind++)
     {
-      char *irix4_cc_needs_this;
-      STRDUP_ALLOCA (irix4_cc_needs_this, argv[optind]);
-      url[i] = irix4_cc_needs_this;
+      char *rewritten = rewrite_shorthand_url (argv[optind]);
+      if (rewritten)
+	{
+	  printf ("Converted %s to %s\n", argv[optind], rewritten);
+	  url[i] = rewritten;
+	}
+      else
+	url[i] = xstrdup (argv[optind]);
     }
   url[i] = NULL;
 
@@ -789,20 +793,28 @@ Can't timestamp and not clobber old files at the same time.\n"));
      process exits.  What we want is to ignore SIGPIPE and just check
      for the return value of write().  */
   signal (SIGPIPE, SIG_IGN);
+#ifdef SIGWINCH
+  signal (SIGWINCH, progress_handle_sigwinch);
+#endif
 #endif /* HAVE_SIGNAL */
 
+#ifdef HAVE_SSL
+  /* Must call this before resolving any URLs because it has the power
+     to disable `https'.  */
+  ssl_init_prng ();
+#endif
+
   status = RETROK;		/* initialize it, just-in-case */
-  recursive_reset ();
   /* Retrieve the URLs from argument list.  */
   for (t = url; *t; t++)
     {
-      char *filename, *redirected_URL;
+      char *filename = NULL, *redirected_URL = NULL;
       int dt;
 
-      status = retrieve_url (*t, &filename, &redirected_URL, NULL, &dt);
-      if (opt.recursive && status == RETROK && (dt & TEXTHTML))
-	status = recursive_retrieve (filename,
-				     redirected_URL ? redirected_URL : *t);
+      if (opt.recursive && url_scheme (*t) != SCHEME_FTP)
+	status = retrieve_tree (*t);
+      else
+	status = retrieve_url (*t, &filename, &redirected_URL, NULL, &dt);
 
       if (opt.delete_after && file_exists_p(filename))
 	{
@@ -850,8 +862,12 @@ Can't timestamp and not clobber old files at the same time.\n"));
     {
       convert_all_links ();
     }
+
   log_close ();
+  for (i = 0; i < nurl; i++)
+    xfree (url[i]);
   cleanup ();
+
 #ifdef DEBUG_MALLOC
   print_malloc_debug_stats ();
 #endif
@@ -861,11 +877,15 @@ Can't timestamp and not clobber old files at the same time.\n"));
     return 1;
 }
 
+#ifdef HAVE_SIGNAL
 /* Hangup signal handler.  When wget receives SIGHUP or SIGUSR1, it
    will proceed operation as usual, trying to write into a log file.
-   If that is impossible, the output will be turned off.  */
+   If that is impossible, the output will be turned off.
+
+   #### It is unsafe to do call libc functions from a signal handler.
+   What we should do is, set a global variable, and have the code in
+   log.c pick it up.  */
 
-#ifdef HAVE_SIGNAL
 static RETSIGTYPE
 redirect_output_signal (int sig)
 {
@@ -878,5 +898,6 @@ redirect_output_signal (int sig)
 	    (sig == SIGUSR1 ? "SIGUSR1" :
 	     "WTF?!")));
   redirect_output (tmp);
+  progress_schedule_redirect ();
 }
 #endif /* HAVE_SIGNAL */
