@@ -36,6 +36,7 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "wget.h"
 #include "utils.h"
 #include "retr.h"
+#include "progress.h"
 #include "url.h"
 #include "recur.h"
 #include "ftp.h"
@@ -43,6 +44,10 @@ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 #include "connect.h"
 #include "hash.h"
 
+#ifdef HAVE_SSL
+# include "gen_sslfunc.h"	/* for ssl_iread */
+#endif
+
 #ifndef errno
 extern int errno;
 #endif
@@ -50,15 +55,50 @@ extern int errno;
 /* See the comment in gethttp() why this is needed. */
 int global_download_count;
 
-void logflush PARAMS ((void));
-
-/* From http.c.  */
-uerr_t http_loop PARAMS ((struct urlinfo *, char **, int *));
 
-/* Flags for show_progress().  */
-enum spflags { SP_NONE, SP_INIT, SP_FINISH };
+static struct {
+  long bytes;
+  long dltime;
+} limit_data;
 
-static int show_progress PARAMS ((long, long, enum spflags));
+static void
+limit_bandwidth_reset (void)
+{
+  limit_data.bytes  = 0;
+  limit_data.dltime = 0;
+}
+
+/* Limit the bandwidth by pausing the download for an amount of time.
+   BYTES is the number of bytes received from the network, DELTA is
+   how long it took to receive them, DLTIME the current download time,
+   TIMER the timer, and ADJUSTMENT the previous.  */
+
+static void
+limit_bandwidth (long bytes, long delta)
+{
+  long expected;
+
+  limit_data.bytes += bytes;
+  limit_data.dltime += delta;
+
+  expected = (long)(1000.0 * limit_data.bytes / opt.limit_rate);
+
+  if (expected > limit_data.dltime)
+    {
+      long slp = expected - limit_data.dltime;
+      if (slp < 200)
+	{
+	  DEBUGP (("deferring a %ld ms sleep (%ld/%ld) until later.\n",
+		   slp, limit_data.bytes, limit_data.dltime));
+	  return;
+	}
+      DEBUGP (("sleeping %ld ms\n", slp));
+      usleep (1000 * slp);
+    }
+
+  limit_data.bytes = 0;
+  limit_data.dltime = 0;
+}
 
 #define MIN(i, j) ((i) <= (j) ? (i) : (j))
 
@@ -85,28 +125,43 @@ static int show_progress PARAMS ((long, long, enum spflags));
    from fd immediately, flush or discard the buffer.  */
 int
 get_contents (int fd, FILE *fp, long *len, long restval, long expected,
-	      struct rbuf *rbuf, int use_expected)
+	      struct rbuf *rbuf, int use_expected, long *elapsed)
 {
   int res = 0;
   static char c[8192];
+  void *progress = NULL;
+  struct wget_timer *timer = wtimer_allocate ();
+  long dltime = 0, last_dltime = 0;
 
   *len = restval;
+
   if (opt.verbose)
-    show_progress (restval, expected, SP_INIT);
+    progress = progress_create (restval, expected);
+
   if (rbuf && RBUF_FD (rbuf) == fd)
     {
+      int sz = 0;
       while ((res = rbuf_flush (rbuf, c, sizeof (c))) != 0)
 	{
-	  if (fwrite (c, sizeof (char), res, fp) < res)
-	    return -2;
-	  if (opt.verbose)
-	    {
-	      if (show_progress (res, expected, SP_NONE))
-		fflush (fp);
-	    }
+	  fwrite (c, sizeof (char), res, fp);
 	  *len += res;
+	  sz += res;
 	}
+      if (sz)
+	fflush (fp);
+      if (ferror (fp))
+	{
+	  res = -2;
+	  goto out;
+	}
+      if (opt.verbose)
+	progress_update (progress, sz, 0);
     }
+
+  if (opt.limit_rate)
+    limit_bandwidth_reset ();
+  wtimer_reset (timer);
+
   /* Read from fd while there is available data.
 
      Normally, if expected is 0, it means that it is not known how
@@ -118,23 +173,37 @@ get_contents (int fd, FILE *fp, long *len, long restval, long expected,
 			    ? MIN (expected - *len, sizeof (c))
 			    : sizeof (c));
 #ifdef HAVE_SSL
-		if (rbuf->ssl!=NULL) {
-		  res = ssl_iread (rbuf->ssl, c, amount_to_read);
-		} else {
-#endif /* HAVE_SSL */
-		  res = iread (fd, c, amount_to_read);
-#ifdef HAVE_SSL
-		}
+      if (rbuf->ssl!=NULL)
+	res = ssl_iread (rbuf->ssl, c, amount_to_read);
+      else
 #endif /* HAVE_SSL */
+	res = iread (fd, c, amount_to_read);
+
       if (res > 0)
 	{
-	  if (fwrite (c, sizeof (char), res, fp) < res)
-	    return -2;
-	  if (opt.verbose)
+	  fwrite (c, sizeof (char), res, fp);
+	  /* Always flush the contents of the network packet.  This
+	     should not be adverse to performance, as the network
+	     packets typically won't be too tiny anyway.  */
+	  fflush (fp);
+	  if (ferror (fp))
 	    {
-	      if (show_progress (res, expected, SP_NONE))
-		fflush (fp);
+	      res = -2;
+	      goto out;
 	    }
+
+	  /* If bandwidth is not limited, one call to wtimer_elapsed
+	     is sufficient.  */
+	  dltime = wtimer_elapsed (timer);
+	  if (opt.limit_rate)
+	    {
+	      limit_bandwidth (res, dltime - last_dltime);
+	      dltime = wtimer_elapsed (timer);
+	      last_dltime = dltime;
+	    }
+
+	  if (opt.verbose)
+	    progress_update (progress, res, dltime);
 	  *len += res;
 	}
       else
@@ -142,142 +211,48 @@ get_contents (int fd, FILE *fp, long *len, long restval, long expected,
     }
   if (res < -1)
     res = -1;
+
+ out:
   if (opt.verbose)
-    show_progress (0, expected, SP_FINISH);
+    progress_finish (progress, dltime);
+  if (elapsed)
+    *elapsed = dltime;
+  wtimer_delete (timer);
+
+  return res;
+}
+
+/* Return a printed representation of the download rate, as
+   appropriate for the speed.  If PAD is non-zero, strings will be
+   padded to the width of 7 characters (xxxx.xx).  */
+char *
+retr_rate (long bytes, long msecs, int pad)
+{
+  static char res[20];
+  static char *rate_names[] = {"B/s", "KB/s", "MB/s", "GB/s" };
+  int units = 0;
+
+  double dlrate = calc_rate (bytes, msecs, &units);
+  sprintf (res, pad ? "%7.2f %s" : "%.2f %s", dlrate, rate_names[units]);
+
   return res;
 }
 
-static void
-print_percentage (long bytes, long expected)
+/* Calculate the download rate and trim it as appropriate for the
+   speed.  Appropriate means that if rate is greater than 1K/s,
+   kilobytes are used, and if rate is greater than 1MB/s, megabytes
+   are used.
+
+   UNITS is zero for B/s, one for KB/s, two for MB/s, and three for
+   GB/s.  */
+double
+calc_rate (long bytes, long msecs, int *units)
 {
-  int percentage = (int)(100.0 * bytes / expected);
-  logprintf (LOG_VERBOSE, "%3d%%", percentage);
-}
-
-/* Show the dotted progress report of file loading.  Called with
-   length and a flag to tell it whether to reset or not.  It keeps the
-   offset information in static local variables.
-
-   Return value: 1 or 0, designating whether any dots have been drawn.
-
-   If the init argument is set, the routine will initialize.
-
-   If the res is non-zero, res/line_bytes lines are skipped
-   (meaning the appropriate number ok kilobytes), and the number of
-   "dots" fitting on the first line are drawn as ','.  */
-static int
-show_progress (long res, long expected, enum spflags flags)
-{
-  static struct wget_timer *timer;
-  static long line_bytes;
-  static long offs, initial_skip;
-  static int ndot, nrow;
-  static long last_timer_value, time_offset;
-  int any_output = 0;
-
-  if (flags == SP_FINISH)
-    {
-      int dot = ndot;
-      char *tmpstr = (char *)alloca (2 * opt.dots_in_line + 1);
-      char *tmpp = tmpstr;
-      time_offset = wtimer_elapsed (timer) - last_timer_value;
-      for (; dot < opt.dots_in_line; dot++)
-	{
-	  if (!(dot % opt.dot_spacing))
-	    *tmpp++ = ' ';
-	  *tmpp++ = ' ';
-	}
-      *tmpp = '\0';
-      logputs (LOG_VERBOSE, tmpstr);
-      if (expected)
-	print_percentage (nrow * line_bytes + ndot * opt.dot_bytes + offs,
-			  expected);
-      logprintf (LOG_VERBOSE, " @%s",
-		 rate (ndot * opt.dot_bytes
-		       + offs - (initial_skip % line_bytes),
-		       time_offset, 1));
-      logputs (LOG_VERBOSE, "\n\n");
-      return 0;
-    }
-
-  /* Temporarily disable flushing.  */
-  opt.no_flush = 1;
-  /* init set means initialization.  If res is set, it also means that
-     the retrieval is *not* done from the beginning.  The part that
-     was already retrieved is not shown again.  */
-  if (flags == SP_INIT)
-    {
-      /* Generic initialization of static variables.  */
-      offs = 0L;
-      ndot = nrow = 0;
-      line_bytes = (long)opt.dots_in_line * opt.dot_bytes;
-      if (!timer)
-	timer = wtimer_allocate ();
-      wtimer_reset (timer);
-      last_timer_value = 0;
-      time_offset = 0;
-      initial_skip = res;
-      if (res)
-	{
-	  if (res >= line_bytes)
-	    {
-	      nrow = res / line_bytes;
-	      res %= line_bytes;
-	      logprintf (LOG_VERBOSE,
-			 _("\n          [ skipping %dK ]"),
-			 (int) ((nrow * line_bytes) / 1024));
-	      ndot = 0;
-	    }
-	}
-      logprintf (LOG_VERBOSE, "\n%5ldK", nrow * line_bytes / 1024);
-    }
-  /* Offset gets incremented by current value.  */
-  offs += res;
-  /* While offset is >= opt.dot_bytes, print dots, taking care to
-     precede every 50th dot with a status message.  */
-  for (; offs >= opt.dot_bytes; offs -= opt.dot_bytes)
-    {
-      if (!(ndot % opt.dot_spacing))
-	logputs (LOG_VERBOSE, " ");
-      any_output = 1;
-      logputs (LOG_VERBOSE, flags == SP_INIT ? "," : ".");
-      ++ndot;
-      if (ndot == opt.dots_in_line)
-	{
-	  time_offset = wtimer_elapsed (timer) - last_timer_value;
-	  last_timer_value += time_offset;
-
-	  ndot = 0;
-	  ++nrow;
-	  if (expected)
-	    print_percentage (nrow * line_bytes, expected);
-	  logprintf (LOG_VERBOSE, " @%s",
-		     rate (line_bytes - (initial_skip % line_bytes),
-			   time_offset, 1));
-	  initial_skip = 0;
-	  logprintf (LOG_VERBOSE, "\n%5ldK", nrow * line_bytes / 1024);
-	}
-    }
-  /* Reenable flushing.  */
-  opt.no_flush = 0;
-  if (any_output)
-    /* Force flush.  */
-    logflush ();
-  return any_output;
-}
-
-/* Print out the appropriate download rate.  Appropriate means that if
-   rate is > 1024 bytes per second, kilobytes are used, and if rate >
-   1024 * 1024 bps, megabytes are used.
-
-   If PAD is non-zero, strings will be padded to the width of 7
-   characters (xxxx.xx).  */
-char *
-rate (long bytes, long msecs, int pad)
-{
-  static char res[15];
   double dlrate;
 
+  assert (msecs >= 0);
+  assert (bytes >= 0);
+
   if (msecs == 0)
     /* If elapsed time is 0, it means we're under the granularity of
        the timer.  This often happens on systems that use time() for
@@ -286,27 +261,53 @@ rate (long bytes, long msecs, int pad)
 
   dlrate = (double)1000 * bytes / msecs;
   if (dlrate < 1024.0)
-    sprintf (res, pad ? "%7.2f B/s" : "%.2f B/s", dlrate);
+    *units = 0;
   else if (dlrate < 1024.0 * 1024.0)
-    sprintf (res, pad ? "%7.2f KB/s" : "%.2f KB/s", dlrate / 1024.0);
+    *units = 1, dlrate /= 1024.0;
   else if (dlrate < 1024.0 * 1024.0 * 1024.0)
-    sprintf (res, pad ? "%7.2f MB/s" : "%.2f MB/s", dlrate / (1024.0 * 1024.0));
+    *units = 2, dlrate /= (1024.0 * 1024.0);
   else
     /* Maybe someone will need this one day.  More realistically, it
        will get tickled by buggy timers. */
-    sprintf (res, pad ? "%7.2f GB/s" : "%.2f GB/s",
-	     dlrate / (1024.0 * 1024.0 * 1024.0));
+    *units = 3, dlrate /= (1024.0 * 1024.0 * 1024.0);
 
-  return res;
+  return dlrate;
 }
 
-#define USE_PROXY_P(u) (opt.use_proxy && getproxy((u)->proto)		\
+static int
+register_redirections_mapper (void *key, void *value, void *arg)
+{
+  const char *redirected_from = (const char *)key;
+  const char *redirected_to   = (const char *)arg;
+  if (0 != strcmp (redirected_from, redirected_to))
+    register_redirection (redirected_from, redirected_to);
+  return 0;
+}
+
+/* Register the redirections that lead to the successful download of
+   this URL.  This is necessary so that the link converter can convert
+   redirected URLs to the local file.  */
+
+static void
+register_all_redirections (struct hash_table *redirections, const char *final)
+{
+  hash_table_map (redirections, register_redirections_mapper, (void *)final);
+}
+
+#define USE_PROXY_P(u) (opt.use_proxy && getproxy((u)->scheme)		\
 			&& no_proxy_match((u)->host,			\
 					  (const char **)opt.no_proxy))
 
-/* Retrieve the given URL.  Decides which loop to call -- HTTP(S), FTP,
-   or simply copy it with file:// (#### the latter not yet
-   implemented!).  */
+/* Maximum number of allowed redirections.  20 was chosen as a
+   "reasonable" value, which is low enough to not cause havoc, yet
+   high enough to guarantee that normal retrievals will not be hurt by
+   the check.  */
+
+#define MAX_REDIRECTIONS 20
+
+/* Retrieve the given URL.  Decides which loop to call -- HTTP, FTP,
+   FTP, proxy, etc.  */
+
 uerr_t
 retrieve_url (const char *origurl, char **file, char **newloc,
 	      const char *refurl, int *dt)
@@ -314,10 +315,13 @@ retrieve_url (const char *origurl, char **file, char **newloc,
   uerr_t result;
   char *url;
   int location_changed, dummy;
-  int local_use_proxy;
+  int use_proxy;
   char *mynewloc, *proxy;
-  struct urlinfo *u;
+  struct url *u;
+  int up_error_code;		/* url parse error code */
+  char *local_file;
   struct hash_table *redirections = NULL;
+  int redirection_count = 0;
 
   /* If dt is NULL, just ignore it.  */
   if (!dt)
@@ -328,82 +332,75 @@ retrieve_url (const char *origurl, char **file, char **newloc,
   if (file)
     *file = NULL;
 
-  u = newurl ();
-  /* Parse the URL. */
-  result = parseurl (url, u, 0);
-  if (result != URLOK)
+  u = url_parse (url, &up_error_code);
+  if (!u)
     {
-      logprintf (LOG_NOTQUIET, "%s: %s.\n", url, uerrmsg (result));
-      freeurl (u, 1);
+      logprintf (LOG_NOTQUIET, "%s: %s.\n", url, url_error (up_error_code));
       if (redirections)
 	string_set_free (redirections);
       xfree (url);
-      return result;
+      return URLERROR;
     }
 
+  if (!refurl)
+    refurl = opt.referer;
+
  redirected:
 
-  /* Set the referer.  */
-  if (refurl)
-    u->referer = xstrdup (refurl);
-  else
-    {
-      if (opt.referer)
-	u->referer = xstrdup (opt.referer);
-      else
-	u->referer = NULL;
-    }
+  result = NOCONERROR;
+  mynewloc = NULL;
+  local_file = NULL;
 
-  local_use_proxy = USE_PROXY_P (u);
-  if (local_use_proxy)
+  use_proxy = USE_PROXY_P (u);
+  if (use_proxy)
     {
-      struct urlinfo *pu = newurl ();
+      struct url *proxy_url;
 
-      /* Copy the original URL to new location.  */
-      memcpy (pu, u, sizeof (*u));
-      pu->proxy = NULL; /* A minor correction :) */
-      /* Initialize u to nil.  */
-      memset (u, 0, sizeof (*u));
-      u->proxy = pu;
-      /* Get the appropriate proxy server, appropriate for the
-	 current protocol.  */
-      proxy = getproxy (pu->proto);
+      /* Get the proxy server for the current scheme.  */
+      proxy = getproxy (u->scheme);
       if (!proxy)
 	{
 	  logputs (LOG_NOTQUIET, _("Could not find proxy host.\n"));
-	  freeurl (u, 1);
+	  url_free (u);
 	  if (redirections)
 	    string_set_free (redirections);
 	  xfree (url);
 	  return PROXERR;
 	}
+
       /* Parse the proxy URL.  */
-      result = parseurl (proxy, u, 0);
-      if (result != URLOK || u->proto != URLHTTP)
+      proxy_url = url_parse (proxy, &up_error_code);
+      if (!proxy_url)
 	{
-	  if (u->proto == URLHTTP)
-	    logprintf (LOG_NOTQUIET, "Proxy %s: %s.\n", proxy, uerrmsg(result));
-	  else
-	    logprintf (LOG_NOTQUIET, _("Proxy %s: Must be HTTP.\n"), proxy);
-	  freeurl (u, 1);
+	  logprintf (LOG_NOTQUIET, _("Error parsing proxy URL %s: %s.\n"),
+		     proxy, url_error (up_error_code));
 	  if (redirections)
 	    string_set_free (redirections);
 	  xfree (url);
 	  return PROXERR;
 	}
-      u->proto = URLHTTP;
+      if (proxy_url->scheme != SCHEME_HTTP)
+	{
+	  logprintf (LOG_NOTQUIET, _("Error in proxy URL %s: Must be HTTP.\n"), proxy);
+	  url_free (proxy_url);
+	  if (redirections)
+	    string_set_free (redirections);
+	  xfree (url);
+	  return PROXERR;
+	}
+
+      result = http_loop (u, &mynewloc, &local_file, refurl, dt, proxy_url);
+      url_free (proxy_url);
     }
-
-  assert (u->proto != URLFILE);	/* #### Implement me!  */
-  mynewloc = NULL;
-
-  if (u->proto == URLHTTP
+  else if (u->scheme == SCHEME_HTTP
 #ifdef HAVE_SSL
-      || u->proto == URLHTTPS
+      || u->scheme == SCHEME_HTTPS
 #endif
       )
-    result = http_loop (u, &mynewloc, dt);
-  else if (u->proto == URLFTP)
+    {
+      result = http_loop (u, &mynewloc, &local_file, refurl, dt, NULL);
+    }
+  else if (u->scheme == SCHEME_FTP)
     {
       /* If this is a redirection, we must not allow recursive FTP
 	 retrieval, so we save recursion to oldrec, and restore it
@@ -413,30 +410,29 @@ retrieve_url (const char *origurl, char **file, char **newloc,
 	opt.recursive = 0;
       result = ftp_loop (u, dt);
       opt.recursive = oldrec;
+
       /* There is a possibility of having HTTP being redirected to
 	 FTP.  In these cases we must decide whether the text is HTML
 	 according to the suffix.  The HTML suffixes are `.html' and
-	 `.htm', case-insensitive.
-
-	 #### All of this is, of course, crap.  These types should be
-	 determined through mailcap.  */
-      if (redirections && u->local && (u->proto == URLFTP ))
+	 `.htm', case-insensitive.  */
+      if (redirections && local_file && u->scheme == SCHEME_FTP)
 	{
-	  char *suf = suffix (u->local);
+	  char *suf = suffix (local_file);
 	  if (suf && (!strcasecmp (suf, "html") || !strcasecmp (suf, "htm")))
 	    *dt |= TEXTHTML;
-	  FREE_MAYBE (suf);
 	}
     }
   location_changed = (result == NEWLOCATION);
   if (location_changed)
     {
       char *construced_newloc;
-      uerr_t newloc_result;
-      struct urlinfo *newloc_struct;
+      struct url *newloc_parsed;
 
       assert (mynewloc != NULL);
 
+      if (local_file)
+	xfree (local_file);
+
       /* The HTTP specs only allow absolute URLs to appear in
 	 redirects, but a ton of boneheaded webservers and CGIs out
 	 there break the rules and use relative URLs, and popular
@@ -446,13 +442,12 @@ retrieve_url (const char *origurl, char **file, char **newloc,
       mynewloc = construced_newloc;
 
       /* Now, see if this new location makes sense. */
-      newloc_struct = newurl ();
-      newloc_result = parseurl (mynewloc, newloc_struct, 1);
-      if (newloc_result != URLOK)
+      newloc_parsed = url_parse (mynewloc, &up_error_code);
+      if (!newloc_parsed)
 	{
-	  logprintf (LOG_NOTQUIET, "%s: %s.\n", mynewloc, uerrmsg (newloc_result));
-	  freeurl (newloc_struct, 1);
-	  freeurl (u, 1);
+	  logprintf (LOG_NOTQUIET, "%s: %s.\n", mynewloc,
+		     url_error (up_error_code));
+	  url_free (u);
 	  if (redirections)
 	    string_set_free (redirections);
 	  xfree (url);
@@ -460,11 +455,11 @@ retrieve_url (const char *origurl, char **file, char **newloc,
 	  return result;
 	}
 
-      /* Now mynewloc will become newloc_struct->url, because if the
+      /* Now mynewloc will become newloc_parsed->url, because if the
          Location contained relative paths like .././something, we
          don't want that propagating as url.  */
       xfree (mynewloc);
-      mynewloc = xstrdup (newloc_struct->url);
+      mynewloc = xstrdup (newloc_parsed->url);
 
       if (!redirections)
 	{
@@ -474,54 +469,77 @@ retrieve_url (const char *origurl, char **file, char **newloc,
 	  string_set_add (redirections, u->url);
 	}
 
-      /* The new location is OK.  Let's check for redirection cycle by
-         peeking through the history of redirections. */
-      if (string_set_contains (redirections, newloc_struct->url))
+      /* The new location is OK.  Check for max. number of
+	 redirections.  */
+      if (++redirection_count > MAX_REDIRECTIONS)
 	{
-	  logprintf (LOG_NOTQUIET, _("%s: Redirection cycle detected.\n"),
-		     mynewloc);
-	  freeurl (newloc_struct, 1);
-	  freeurl (u, 1);
+	  logprintf (LOG_NOTQUIET, _("%d redirections exceeded.\n"),
+		     MAX_REDIRECTIONS);
+	  url_free (newloc_parsed);
+	  url_free (u);
 	  if (redirections)
 	    string_set_free (redirections);
 	  xfree (url);
 	  xfree (mynewloc);
 	  return WRONGCODE;
 	}
-      string_set_add (redirections, newloc_struct->url);
+
+      /*Check for redirection cycle by
+         peeking through the history of redirections. */
+      if (string_set_contains (redirections, newloc_parsed->url))
+	{
+	  logprintf (LOG_NOTQUIET, _("%s: Redirection cycle detected.\n"),
+		     mynewloc);
+	  url_free (newloc_parsed);
+	  url_free (u);
+	  if (redirections)
+	    string_set_free (redirections);
+	  xfree (url);
+	  xfree (mynewloc);
+	  return WRONGCODE;
+	}
+      string_set_add (redirections, newloc_parsed->url);
 
       xfree (url);
       url = mynewloc;
-      freeurl (u, 1);
-      u = newloc_struct;
+      url_free (u);
+      u = newloc_parsed;
       goto redirected;
     }
 
-  if (u->local)
+  if (local_file)
     {
       if (*dt & RETROKF)
 	{
-	  register_download (url, u->local);
+	  register_download (u->url, local_file);
+	  if (redirections)
+	    register_all_redirections (redirections, u->url);
 	  if (*dt & TEXTHTML)
-	    register_html (url, u->local);
+	    register_html (u->url, local_file);
 	}
     }
 
   if (file)
-    {
-      if (u->local)
-	*file = xstrdup (u->local);
-      else
-	*file = NULL;
-    }
-  freeurl (u, 1);
-  if (redirections)
-    string_set_free (redirections);
-
-  if (newloc)
-    *newloc = url;
+    *file = local_file ? local_file : NULL;
   else
-    xfree (url);
+    FREE_MAYBE (local_file);
+
+  url_free (u);
+
+  if (redirections)
+    {
+      string_set_free (redirections);
+      if (newloc)
+	*newloc = url;
+      else
+	xfree (url);
+    }
+  else
+    {
+      if (newloc)
+	*newloc = NULL;
+      xfree (url);
+    }
 
   ++global_download_count;
 
@@ -537,27 +555,30 @@ uerr_t
 retrieve_from_file (const char *file, int html, int *count)
 {
   uerr_t status;
-  urlpos *url_list, *cur_url;
+  struct urlpos *url_list, *cur_url;
 
-  url_list = (html ? get_urls_html (file, NULL, FALSE, NULL)
+  url_list = (html ? get_urls_html (file, NULL, NULL)
 	      : get_urls_file (file));
   status = RETROK;             /* Suppose everything is OK.  */
   *count = 0;                  /* Reset the URL count.  */
-  recursive_reset ();
+
   for (cur_url = url_list; cur_url; cur_url = cur_url->next, ++*count)
     {
-      char *filename, *new_file;
+      char *filename = NULL, *new_file = NULL;
       int dt;
 
+      if (cur_url->ignore_when_downloading)
+	continue;
+
       if (downloaded_exceeds_quota ())
 	{
 	  status = QUOTEXC;
 	  break;
 	}
-      status = retrieve_url (cur_url->url, &filename, &new_file, NULL, &dt);
-      if (opt.recursive && status == RETROK && (dt & TEXTHTML))
-	status = recursive_retrieve (filename, new_file ? new_file
-				                        : cur_url->url);
+      if (opt.recursive && cur_url->url->scheme != SCHEME_FTP)
+	status = retrieve_tree (cur_url->url->url);
+      else
+	status = retrieve_url (cur_url->url->url, &filename, &new_file, NULL, &dt);
 
       if (filename && opt.delete_after && file_exists_p (filename))
 	{
@@ -632,6 +653,10 @@ sleep_between_retrievals (int count)
 {
   static int first_retrieval = 1;
 
+  if (first_retrieval && opt.random_wait)
+    /* --random-wait uses the RNG, so seed it. */
+    srand (time (NULL));
+
   if (!first_retrieval && (opt.wait || opt.waitretry))
     {
       if (opt.waitretry && count > 1)
@@ -645,8 +670,24 @@ sleep_between_retrievals (int count)
 	    sleep (opt.waitretry);
 	}
       else if (opt.wait)
-	/* Otherwise, check if opt.wait is specified.  If so, sleep.  */
-	sleep (opt.wait);
+	{
+	  /* Otherwise, check if opt.wait is specified.  If so, sleep.  */
+	  if (count > 1 || !opt.random_wait)
+	    sleep (opt.wait);
+	  else
+	    {
+	      int waitmax = 2 * opt.wait;
+	      /* This is equivalent to rand() % waitmax, but uses the
+		 high-order bits for better randomness.  */
+	      int waitsecs = (double)waitmax * rand () / (RAND_MAX + 1.0);
+
+	      DEBUGP (("sleep_between_retrievals: norm=%ld,fuzz=%ld,sleep=%d\n",
+		       opt.wait, waitsecs - opt.wait, waitsecs));
+
+	      if (waitsecs)
+		sleep (waitsecs);
+	    }
+	}
     }
   if (first_retrieval)
     first_retrieval = 0;
