@@ -44,18 +44,19 @@ struct wget_timer;
 char *time_str PARAMS ((time_t *));
 char *datetime_str PARAMS ((time_t *));
 
-const char *uerrmsg PARAMS ((uerr_t));
-
 #ifdef DEBUG_MALLOC
 void print_malloc_debug_stats ();
 #endif
 
+char *xstrdup_lower PARAMS ((const char *));
+int count_char PARAMS ((const char *, char));
+
 char *strdupdelim PARAMS ((const char *, const char *));
 char **sepstring PARAMS ((const char *));
 int frontcmp PARAMS ((const char *, const char *));
 char *pwd_cuserid PARAMS ((char *));
 void fork_to_background PARAMS ((void));
-void path_simplify PARAMS ((char *));
+int path_simplify PARAMS ((char *));
 
 void touch PARAMS ((const char *, time_t));
 int remove_link PARAMS ((const char *));
@@ -63,10 +64,12 @@ int file_exists_p PARAMS ((const char *));
 int file_non_directory_p PARAMS ((const char *));
 int make_directory PARAMS ((const char *));
 char *unique_name PARAMS ((const char *));
+char *file_merge PARAMS ((const char *, const char *));
 
 int acceptable PARAMS ((const char *));
 int accdir PARAMS ((const char *s, enum accd));
 char *suffix PARAMS ((const char *s));
+int match_tail PARAMS ((const char *, const char *));
 
 char *read_whole_line PARAMS ((FILE *));
 struct file_memory *read_file PARAMS ((const char *));
@@ -99,4 +102,6 @@ long wtimer_granularity PARAMS ((void));
 
 char *html_quote_string PARAMS ((const char *));
 
+int determine_screen_width PARAMS ((void));
+
 #endif /* UTILS_H */
